import numpy as np

class NoSmooth:
    def __init__(self, mu):
        self.mu = mu
    def val(self, x):
        return np.zeros_like(x)
    def grad(self, x):
        return np.zeros_like(x)
    def hess(self, x):
        return np.finfo(float).eps * np.ones_like(x)

def get_Mg(Mh, nu, mu, n):
    if Mh < 0:
        raise ValueError("Mh must be nonnegative.")
    if mu <= 0:
        raise ValueError("mu must be positive.")
    if 0 < nu <= 3:
        return n ** ((3 - nu) / 2) * mu ** (nu / 2 - 2) * Mh
    elif nu > 3:
        return mu ** (4 - 3 * nu / 2) * Mh
    else:
        raise ValueError("nu must be positive.")


import numpy as np
import jax
from pyscsopt.regularizers.regularizers import get_reg

class Problem:
    def __init__(self, x0, f, lam, L=None, x=None, C_set=None, P=None, re=None, grad_fx=None, hess_fx=None, jac_yx=None, grad_fy=None, hess_fy=None, name=None, A=None, y=None, out_fn=None):
        self.x0 = x0
        self.f = f
        self.lam = lam
        self.L = L
        self.x = x if x is not None else np.zeros_like(x0)
        self.C_set = C_set
        self.P = P
        self.re = re
        self.grad_fx = grad_fx
        self.hess_fx = hess_fx
        self.jac_yx = jac_yx
        self.grad_fy = grad_fy
        self.hess_fy = hess_fy
        self.name = name
        self.A = A
        self.y = y
        self.out_fn = out_fn

    def get_reg(self, x, reg_name):
        return get_reg(self, x, reg_name)


import jax.numpy as jnp

def indbox_f(x, lb, ub):
    if jnp.any(x < lb) or jnp.any(x > ub):
        return jnp.inf
    else:
        return 0.0

def get_reg(model, x, reg_name):
    if reg_name == "l1":
        return model.lam * jnp.sum(jnp.abs(x))
    elif reg_name == "l2":
        return model.lam * jnp.sum(x ** 2)
    elif reg_name == "indbox":
        lb, ub = model.C_set[0], model.C_set[1]
        return indbox_f(x, lb, ub)
    elif reg_name == "gl":
        if len(model.lam) != 2:
            raise ValueError("Please provide a tuple or list with exactly two entries for lam, e.g. [lam1, lam2]")
        P = model.P
        Px = P.matrix @ x
        lam1, lam2 = model.lam[0], model.lam[1]
        return lam2 * P.Lasso_fz(Px) + lam1 * jnp.sum(jnp.abs(x))
    else:
        raise ValueError("reg_name not valid.")


import numpy as np
from datetime import datetime

class Solution:
    def __init__(self, x, obj, fval, fvaltest, rel, objrel, metricvals, times, epochs, model, grad_norms=None):
        self.x = x
        self.obj = obj
        self.fval = fval
        self.fvaltest = fvaltest
        self.rel = rel
        self.objrel = objrel
        self.metricvals = metricvals
        self.times = times
        self.epochs = epochs
        self.model = model
        self.grad_norms = grad_norms

class Options:
    def __init__(self, metrics=None, alpha=None, batch_size=None, slice_samples=False, shuffle_batch=True, max_epoch=1, max_iter=None, comm_rounds=100, local_max_iter=None, x_tol=1e-10, f_tol=1e-10, verbose=1, patience=3):
        self.metrics = metrics
        self.alpha = alpha
        self.batch_size = batch_size
        self.slice_samples = slice_samples
        self.shuffle_batch = shuffle_batch
        self.max_epoch = max_epoch
        self.max_iter = max_iter
        self.comm_rounds = comm_rounds
        self.local_max_iter = local_max_iter
        self.x_tol = x_tol
        self.f_tol = f_tol
        self.verbose = verbose
        self.patience = patience

def iter_step(method, model, reg_name, hmu, As, x, x_prev, ys, Cmat, iter):
    return method.step(model, reg_name, hmu, As, x, x_prev, ys, Cmat, iter)

def iterate(method, model, reg_name, hmu, metrics=None, alpha=None, batch_size=None, slice_samples=False, shuffle_batch=True, max_epoch=1000, comm_rounds=100, local_max_iter=None, x_tol=1e-10, f_tol=1e-10, verbose=1, patience=5):
    opt = Options(
        metrics=metrics,
        alpha=alpha,
        batch_size=batch_size,
        slice_samples=slice_samples,
        shuffle_batch=shuffle_batch,
        max_epoch=1 if local_max_iter is not None else max_epoch,
        comm_rounds=comm_rounds,
        local_max_iter=local_max_iter,
        x_tol=x_tol,
        f_tol=f_tol,
        verbose=verbose,
        patience=patience
    )
    return optim_loop(method, model, reg_name, hmu, opt)

def optim_loop(method, model, reg_name, hmu, opt):
    is_generic = (model.A is None or model.y is None)
    x = model.x0.copy()
    x_prev = x.copy()
    epochs = 0
    objs, fvals, rel_errors, objrels, times = [], [], [], [], []
    grad_norms = []
    t0 = datetime.now()
    # Print optimization header
    if opt.verbose > 0:
        print("\n========== Optimization Started ==========")
        print(f"Method:      {getattr(method, 'name', str(method))}")
        print(f"Model:       {getattr(model, 'name', type(model).__name__)}")
        print(f"Regularizer: {reg_name}")
        print(f"Max Epochs:  {opt.max_epoch}")
        print(f"x_tol:       {opt.x_tol:.1e} | f_tol: {opt.f_tol:.1e} | Patience: {opt.patience}")
        print(f"Start Time:  {t0.strftime('%Y-%m-%d %H:%M:%S')}")
        print("-"*50)
        header_fmt = "{:<5s} | {:>12s} | {:>12s} | {:>10s} | {:>10s} | {:>12s}"
        row_fmt =    "{:<5d} | {:>12.4e} | {:>12.4e} | {:>10.4e} | {:>10.2e} | {:>12.2e}"
        print(header_fmt.format('Epoch', 'obj', 'fval', 'rel_err', 'objrel', 'grad_norm'))
        print("-"*77)
    # initialize the method with the starting point x (important for BFGS/Hessian-based methods)
    if method is not None and hasattr(method, 'init'):
        method.init(x)
    if is_generic:
        f = lambda x: model.f(x)
    else:
        f = lambda x: model.f(model.A, model.y, x)
    n = x.shape[0]
    patience_counter = 0
    best_fval = None
    converged_reason = None
    for epoch in range(opt.max_epoch):
        fval = f(x)
        obj = fval + model.get_reg(x, reg_name)
        fvals.append(fval)
        objs.append(obj)
        rel_error = np.linalg.norm(x - x_prev)/max(np.linalg.norm(x_prev), 1)
        rel_errors.append(rel_error)
        objrel = abs(obj - (f(x_prev) + model.get_reg(x_prev, reg_name)))/max(abs((f(x_prev) + model.get_reg(x_prev, reg_name))), 1)
        objrels.append(objrel)
        times.append((datetime.now() - t0).total_seconds())
        if reg_name == "gl" and hasattr(model, "P") and hasattr(model.P, "Cmat"):
            Cmat = model.P.Cmat
        else:
            Cmat = np.eye(n)
        if best_fval is None or fval < best_fval:
            best_fval = fval
            patience_counter = 0
        else:
            patience_counter += 1
        if patience_counter >= opt.patience:
            converged_reason = f"Patience level {opt.patience} reached at epoch {epoch+1}, stopping early."
            if opt.verbose > 0:
                print(f"Patience level {opt.patience} reached at epoch {epoch+1}, stopping early.")
            break
        if patience_counter < opt.patience - 1:
            if is_generic:
                x_new, grad_norm = iter_step(method, model, reg_name, hmu, None, x, x_prev, None, Cmat, epoch)
            else:
                x_new, grad_norm = iter_step(method, model, reg_name, hmu, model.A, x, x_prev, model.y, Cmat, epoch)
            grad_norms.append(grad_norm)
            if opt.verbose > 0:
                # Only print titles if not about to break
                will_break = (
                    np.linalg.norm(x_new - x) < opt.x_tol*max(np.linalg.norm(x), 1)
                    or grad_norm < opt.x_tol * max(1, np.linalg.norm(x))
                )
                if (epoch+1) % 7 == 1 and epoch != 0 and not will_break:
                    print("-"*77)
                    print(header_fmt.format('Epoch', 'obj', 'fval', 'rel_err', 'objrel', 'grad_norm'))
                    print("-"*77)
                print(row_fmt.format(epoch+1, obj, fval, rel_error, objrel, grad_norm))
            if (
                np.linalg.norm(x_new - x) < opt.x_tol*max(np.linalg.norm(x), 1)
                or grad_norm < opt.x_tol * max(1, np.linalg.norm(x))
                ):
                if np.linalg.norm(x_new - x) < opt.x_tol*max(np.linalg.norm(x), 1):
                    converged_reason = f"Converged: ||x_new - x|| < x_tol at epoch {epoch+1}."
                elif grad_norm < opt.x_tol * max(1, np.linalg.norm(x)):
                    converged_reason = f"Converged: grad_norm < x_tol at epoch {epoch+1}."
                break
            x_prev = x.copy()
            x = x_new.copy()
            epochs += 1
        else:
            # grad_norms.append(np.nan)
            converged_reason = f"Stopped at epoch {epoch+1} due to patience policy."
            break
    t1 = datetime.now()
    if opt.verbose > 0:
        print("-"*77)
        print(f"Optimization finished at {t1.strftime('%Y-%m-%d %H:%M:%S')}")
        print(f"Total epochs: {epochs}")
        print(f"Elapsed time: {(t1-t0).total_seconds():.2f} seconds")
        print(f"Best fval:    {np.min(fvals):.6e}")
        print(f"Best obj:     {np.min(objs):.6e}")
        print(f"Final rel_error: {rel_errors[-1]:.2e}")
        print(f"Final objrel:   {objrels[-1]:.2e}")
        if converged_reason:
            print(f"Convergence reason: {converged_reason}")
        print("========== Optimization Complete =========\n")
    return Solution(x, objs, fvals, None, rel_errors, objrels, grad_norms, times, epochs, model, grad_norms=grad_norms)


import numpy as np
from pyscsopt.utils.prox_reg_utils import bounds_sanity_check

def logexp_smooth2_indbox(x, mu=1.0, lb=0.0, ub=1.0):
    n = x.shape[0]
    a, b = bounds_sanity_check(n, lb, ub)
    out = np.zeros(n)
    for i in range(n):
        if x[i] <= a[i] + mu:
            out[i] += (a[i] - x[i] + 3*mu)*(a[i] - x[i] + mu)/(2*mu)
        elif x[i] >= b[i] - mu:
            out[i] += (x[i] - b[i] + 3*mu)*(x[i] - b[i] + mu)/(2*mu)
        if x[i] < a[i]:
            out[i] += mu * (np.log(mu) - np.log(x[i] - a[i]))
        elif x[i] > b[i]:
            out[i] += mu * (np.log(mu) - np.log(b[i] - x[i]))
    return out

def logexp_smooth2_grad_indbox(x, mu=1.0, lb=0.0, ub=1.0):
    n = x.shape[0]
    a, b = bounds_sanity_check(n, lb, ub)
    out = np.zeros(n)
    for i in range(n):
        if x[i] <= a[i] + mu:
            out[i] += (x[i] - a[i] - 2*mu)/mu
        elif x[i] >= b[i] - mu:
            out[i] += (x[i] - b[i] + 2*mu)/mu
        if x[i] < a[i]:
            out[i] += mu/(a[i] - x[i])
        elif x[i] > b[i]:
            out[i] += -mu/(b[i] - x[i])
    return out

def logexp_smooth2_hess_indbox(x, mu=1.0, lb=0.0, ub=1.0):
    n = x.shape[0]
    a, b = bounds_sanity_check(n, lb, ub)
    out = np.zeros(n)
    for i in range(n):
        if x[i] <= a[i] + mu:
            out[i] += 1/mu
        elif x[i] >= b[i] - mu:
            out[i] += 1/mu
        if x[i] < a[i]:
            out[i] += mu/(a[i] - x[i])**2
        elif x[i] > b[i]:
            out[i] += mu/(b[i] - x[i])**2
    return out

class LogExpSmootherIndBox:
    def __init__(self, C_set, mu):
        self.mu = mu
        self.lb = C_set[0]
        self.ub = C_set[1]
        self.Mh = 1.0
        self.nu = 2.0
    def val(self, x):
        return logexp_smooth2_indbox(x, self.mu, self.lb, self.ub)
    def grad(self, Cmat, x):
        return logexp_smooth2_grad_indbox(x, self.mu, self.lb, self.ub)
    def hess(self, Cmat, x):
        return logexp_smooth2_hess_indbox(x, self.mu, self.lb, self.ub)


import numpy as np

def mean_square_error(xtrue, xpred):
    return np.mean((xtrue - xpred) ** 2)

def linesearch(x, d, f, grad_f, max_iter=50, tol=1e-6):
    alpha = 1.0
    rho = 0.5
    c = 1e-4
    for _ in range(max_iter):
        if f(x + alpha * d) <= f(x) + c * alpha * np.dot(grad_f(x), d):
            break
        if np.abs(alpha) < tol:
            break
        alpha *= rho
    return alpha

# compute inverse of the Barzilai-Borwein step size to estimate alpha in the paper
## see https://en.wikipedia.org/wiki/Barzilai-Borwein_method
def inv_BB_step(x, x_prev, gradx, gradx_prev):
    delta = x - x_prev
    gamma = gradx - gradx_prev
    denom = np.dot(delta, gamma)
    if denom == 0:
        return 1.0
    L_est = np.dot(gamma, gamma) / denom
    return L_est


import numpy as np
import scipy.sparse as sp
import jax.numpy as jnp
import scipy.linalg

L_INF_CACHE = -1e32
U_INF_CACHE = 1e32

def bounds_sanity_check(n, lb, ub):
    # Accept scalar or array for lb/ub, and broadcast as needed
    if isinstance(lb, (int, float)):
        lb = np.full(n, lb)
    if isinstance(ub, (int, float)):
        ub = np.full(n, ub)
    lb = np.asarray(lb)
    ub = np.asarray(ub)
    na = lb.size
    nb = ub.size
    if na == n and nb == n:
        a = lb
        b = ub
    else:
        raise ValueError("Lengths of the bounds do not match that of the variable.")
    a = np.where(a == -np.inf, L_INF_CACHE, a)
    b = np.where(b == np.inf, U_INF_CACHE, b)
    return a, b

def twonorm(z, g_start, g_end):
    return jnp.linalg.norm(z[g_start:g_end+1])

def fz(z, ind, grpNUM):
    fz_val = 0.0
    for j in range(grpNUM):
        g_start = int(ind[0, j])
        g_end = int(ind[1, j])
        nrmval = twonorm(z, g_start, g_end)
        fz_val += ind[2, j] * nrmval
    return fz_val

def ProjL2(x, lam, h, inds, grpNUM):
    Px = np.zeros_like(x)
    ind = inds.flatten(order='F')
    for j in range(grpNUM):
        beta_g = lam * ind[2+3*j]
        g_start = int(ind[3*j])
        g_end = int(ind[1+3*j])
        nrmval = twonorm(x / h, g_start, g_end)
        for k in range(g_start, g_end+1):
            Px[k] = x[k] * min(beta_g / (h[k] * nrmval), 1)
    return Px

def ProxL2(x, lam, h, inds, grpNUM):
    Px = np.zeros_like(x)
    ind = inds.flatten(order='F')
    for j in range(grpNUM):
        beta_g = lam * ind[2+3*j]
        g_start = int(ind[3*j])
        g_end = int(ind[1+3*j])
        nrmval = twonorm(x, g_start, g_end)
        for k in range(g_start, g_end+1):
            Px[k] = x[k] * max(1 - beta_g / (h[k] * nrmval), 0)
    return Px

def get_Cmat(ind, grpSIZES, n):
    grpNUM = ind.shape[1]
    g_start = ind[0, :]
    g_end = ind[1, :]
    T = np.zeros((grpNUM, n), dtype=bool)
    for g in range(grpNUM):
        T[g, g_start[g]:g_end[g]+1] = 1
    Tw = ind[2, :]
    V, K = T.shape
    SV = np.sum(grpSIZES)
    J = np.zeros(SV, dtype=int)
    W = np.zeros(SV)
    idx = 0
    for v in range(V):
        idxs = np.where(T[v, :])[0]
        J[idx:idx+len(idxs)] = idxs
        W[idx:idx+len(idxs)] = Tw[v]
        idx += len(idxs)
    C = sp.coo_matrix((W, (np.arange(SV), J)), shape=(SV, K)).tocsc()
    return C

class get_P:
    def __init__(self, grpNUM, grpSIZES, ntotal, ind, G, matrix, Cmat, Pi, tau, times, trans, ProjL2, ProxL2, Lasso_fz):
        self.grpNUM = grpNUM
        self.grpSIZES = grpSIZES
        self.ntotal = ntotal
        self.ind = ind
        self.G = G
        self.matrix = matrix
        self.Cmat = Cmat
        self.Pi = Pi
        self.tau = tau
        self.times = times
        self.trans = trans
        self.ProjL2 = ProjL2
        self.ProxL2 = ProxL2
        self.Lasso_fz = Lasso_fz

def get_P_func(n, G, ind):
    grpNUM = ind.shape[1]
    grpSIZES = (ind[1, :] - ind[0, :] + 1)
    ntotal = np.sum(grpSIZES)
    row = np.arange(ntotal)
    col = G
    data = np.ones(ntotal)
    Pmat = sp.coo_matrix((data, (row, col)), shape=(ntotal, n)).tocsc()
    Pmat_dense = Pmat.toarray() if hasattr(Pmat, 'toarray') else np.array(Pmat)
    Pmat_jax = jnp.array(Pmat_dense)
    def P_i(i):
        tmp = grpSIZES[i]
        I = np.arange(tmp)
        J = G[ind[0, i]:ind[1, i]+1]
        V = np.ones(tmp)
        Pi = sp.coo_matrix((V, (I, J)), shape=(tmp, n)).tocsc()
        Pi_dense = Pi.toarray() if hasattr(Pi, 'toarray') else np.array(Pi)
        return jnp.array(Pi_dense)
    Cmat = get_Cmat(ind, grpSIZES, n)
    Cmat_dense = Cmat.toarray() if hasattr(Cmat, 'toarray') else np.array(Cmat)
    Cmat_jax = jnp.array(Cmat_dense)
    return get_P(
        grpNUM,
        grpSIZES,
        ntotal,
        ind,
        G,
        Pmat_jax,
        Cmat_jax,
        P_i,
        1.0,
        lambda x: Pmat_jax @ x,
        lambda y: Pmat_jax.T @ y,
        lambda z, c1, h: ProjL2(z, c1, h, ind, grpNUM),
        lambda z, c1, h: ProxL2(z, c1, h, ind, grpNUM),
        lambda z: fz(z, ind, grpNUM)
    )

def make_group_lasso_problem(m, n, grpsize=5, p_active=0.1, noise_std=0.1, seed=None, group_weights=1.0, use_const_grpsize=True, corr=0.5):
    """
    Utility to generate a synthetic group lasso regression problem.
    According to the paper "GAP Safe Screening Rules for Sparse-Group Lasso". See https://github.com/EugeneNdiaye/GAPSAFE_SGL
    Args:
        m: number of samples
        n: number of features
        grpsize: size of each group (int or list of ints)
        p_active: proportion of groups that are active (float in (0,1])
        noise_std: standard deviation of noise
        seed: random seed
        group_weights: scalar or list of group weights
        use_const_grpsize: if True, all groups have size grpsize
        corr: correlation for Toeplitz covariance
    Returns:
        A, y, x_true, x0, groups, ind, P
    """
    rng = np.random.default_rng(seed)
    # Group structure
    if use_const_grpsize:
        n_groups = n//grpsize
        groups = [range(i*grpsize, min((i+1)*grpsize, n)) for i in range(n_groups)]
    else:
        splits = np.sort(rng.choice(np.arange(1, n), size=n//grpsize-1, replace=False))
        groups = np.split(np.arange(n), splits)
        n_groups = len(groups)
    # Group weights
    if np.isscalar(group_weights):
        group_weights = [group_weights] * n_groups
    # True coefficients: only a proportion p_active of groups are nonzero
    x_true = np.zeros(n)
    n_active_groups = max(1, int(np.ceil(n_groups * p_active)))
    active_groups = rng.choice(n_groups, size=n_active_groups, replace=False)
    for g in active_groups:
        g_idx = groups[g]
        # For each active group, only a proportion p_active of features are nonzero
        n_features = len(g_idx)
        n_active_features = max(1, int(np.ceil(n_features * p_active)))
        selected_features = rng.choice(g_idx, size=n_active_features, replace=False)
        s = 2 * rng.random(n_active_features) - 1
        u = rng.random(n_active_features)
        x_true[selected_features] = np.sign(s) * (10 * u + (1 - u) * 0.5)
    # Toeplitz covariance for correlated features
    cov = scipy.linalg.toeplitz(corr ** np.arange(n))
    A = rng.multivariate_normal(np.zeros(n), cov, size=m)
    y = A @ x_true + noise_std * rng.normal(size=m)
    x0 = np.zeros(n)
    # ind matrix for group lasso
    ind = np.zeros((3, n_groups), dtype=int)
    for j, g in enumerate(groups):
        ind[0, j] = g.start if isinstance(g, range) else g[0]
        ind[1, j] = g.stop - 1 if isinstance(g, range) else g[-1]
        ind[2, j] = group_weights[j]
    G = np.arange(n)
    P = get_P_func(n, G, ind)
    return A, y, x_true, x0, groups, ind, P



import numpy as np
import jax.numpy as jnp
import pytest
from pyscsopt.problems import Problem
from pyscsopt.algorithms.prox_n_score import ProxNSCORE
from pyscsopt.algorithms.prox_ggn_score import ProxGGNSCORE
from pyscsopt.regularizers.phuber_smooth import PHuberSmootherL1L2, PHuberSmootherIndBox
from pyscsopt.regularizers.exponential_smooth import ExponentialSmootherIndBox
from pyscsopt.algorithms.iterate import iterate

TOL = 1e-6

@pytest.mark.parametrize("reg_name", ["l1", "l2"])
def test_proximal_newton_score_regression(reg_name):
    A = jnp.array([[-0.560501, 0.0], [0.0, 1.85278], [-0.0192918, -0.827763], [0.128064, 0.110096], [0.0, -0.251176]])
    y = jnp.array([-1, -1, -1, 1, -1])
    x0 = jnp.array([0.5908446386657102, 0.7667970365022592])
    lam = 1
    mu = 1
    def f_reg(A, y, x):
        return 1/5 * jnp.sum(jnp.log(1 + jnp.exp(-y * (A @ x))))
    model = Problem(x0=x0, f=f_reg, lam=lam, A=A, y=y)
    hmu = PHuberSmootherL1L2(mu)
    sol = iterate(ProxNSCORE(), model, reg_name, hmu, max_epoch=50)
    assert sol.epochs+1 >= 1
    # assert sol.rel[-1] <= TOL
    assert sol.objrel[-1] <= TOL

@pytest.mark.parametrize("reg_name", ["l1", "l2"])
def test_proximal_ggn_score_regression(reg_name):
    A = jnp.array([[-0.560501, 0.0], [0.0, 1.85278], [-0.0192918, -0.827763], [0.128064, 0.110096], [0.0, -0.251176]])
    y = jnp.array([-1, -1, -1, 1, -1])
    x0 = jnp.array([0.5908446386657102, 0.7667970365022592])
    lam = 1
    mu = 1.0
    def f_reg(A, y, x):
        return 1/5 * jnp.sum(jnp.log(1 + jnp.exp(-y * (A @ x))))
    def Mfunc(A, x):
        return 1 / (1 + jnp.exp(-A @ x))
    model = Problem(x0=x0, f=f_reg, lam=lam, A=A, y=y, out_fn=Mfunc)
    hmu = PHuberSmootherL1L2(mu)
    sol = iterate(ProxGGNSCORE(), model, reg_name, hmu, max_epoch=50)
    assert sol.epochs+1 >= 1
    # assert sol.rel[-1] <= TOL
    assert sol.objrel[-1] <= TOL

def test_phuber_and_exp_indbox():
    A = jnp.array([
        [1.53976, 0.201833, 0.433995, 0.156497, 0.180124],
        [0.201833, 2.37257, -0.0594941, -0.671533, 0.0739676],
        [0.433995, -0.0594941, 3.15025, 0.808797, 0.954656],
        [0.156497, -0.671533, 0.808797, 2.74361, 0.5621],
        [0.180124, 0.0739676, 0.954656, 0.5621, 1.76141]
    ])
    y = jnp.array([0.8673472019512456, -0.9017438158568171, -0.4944787535042339, -0.9029142938652416, 0.8644013132535154])
    x0 = jnp.array([-2.07754990163271, -2.311005948690538, -0.25157276401631606, -0.8858618022602884, 1.3116613046047525])
    x_star = jnp.array([-0.7139006111210786, 0.642716661564418, 0.3684773651494535, 0.5890487798472874, -0.8324174178513779])
    lam = 1.0e-4
    mu = 0.6
    lb = -1.0
    ub = 1.0
    def f_qp(A, y, x):
        return 0.5 * x @ (A @ x) + y @ x
    model_phuber = Problem(x0=x0, f=f_qp, lam=lam, A=A, y=y, C_set=[lb, ub], x=x_star)
    hmu_phuber = PHuberSmootherIndBox(lb, ub, mu)
    sol_p = iterate(ProxNSCORE(), model_phuber, "indbox", hmu_phuber, max_epoch=50, alpha=0.8)
    assert sol_p.epochs+1 >= 1
    # assert sol_p.rel[-1] <= 1e-3
    assert sol_p.objrel[-1] <= 1e-3
    model_exp = Problem(x0=x0, f=f_qp, lam=lam, A=A, y=y, C_set=[lb, ub], x=x_star)
    hmu_exp = ExponentialSmootherIndBox((lb, ub), mu)
    sol_e = iterate(ProxNSCORE(), model_exp, "indbox", hmu_exp, max_epoch=50, alpha=1.0)
    assert sol_e.epochs+1 >= 1
    # assert sol_e.rel[-1] <= 1e-3
    assert sol_e.objrel[-1] <= 1e-3



import numpy as np
import jax
import jax.numpy as jnp
from pyscsopt.utils.utils import linesearch, inv_BB_step
from pyscsopt.regularizers.smoothing import get_Mg
import pyscsopt.prox.prox_operators as prox_ops

class ProxLQNSCORE:
    def __init__(self, ss_type=1, use_prox=True, name="prox-lbfgsscore", label="Prox-LBFGS-SCORE", m=10):
        self.ss_type = ss_type
        self.use_prox = use_prox
        self.name = name
        self.label = label
        self.m = m  # memory size
        self.s_list = []  # s = x_{k+1} - x_k
        self.y_list = []  # y = grad_{k+1} - grad_k
        self.H0 = None  # initial Hessian scaling (scalar or identity)

    def init(self, x):
        n = x.shape[0]
        self.s_list = []
        self.y_list = []
        self.H0 = np.eye(n)
        return self

    def two_loop_recursion(self, grad):
        q = grad.copy()
        alpha = []
        rho = []
        for s, y in zip(reversed(self.s_list), reversed(self.y_list)):
            rho_i = 1.0 / np.dot(y, s)
            alpha_i = rho_i * np.dot(s, q)
            q = q - alpha_i * y
            alpha.append(alpha_i)
            rho.append(rho_i)
        r = self.H0 @ q
        for i in range(len(self.s_list)):
            s = self.s_list[i]
            y = self.y_list[i]
            rho_i = rho[-(i+1)]
            alpha_i = alpha[-(i+1)]
            beta = rho_i * np.dot(y, r)
            r = r + s * (alpha_i - beta)
        return -r

    def step(self, model, reg_name, hmu, As, x, x_prev, ys, Cmat, iter):
        lam = model.lam[0] if hasattr(model.lam, '__len__') and len(model.lam) > 1 else model.lam
        gr = hmu.grad(Cmat, x)
        lam_gr = lam * gr
        Hr_diag = hmu.hess(Cmat, x)
        is_generic = (model.A is None or model.y is None)
        if is_generic:
            obj = lambda x_: model.f(x_) + model.get_reg(x_, reg_name)
            grad_f = model.grad_fx if hasattr(model, 'grad_fx') and model.grad_fx is not None else lambda x_: jnp.array(jax.grad(obj)(x_))
        else:
            obj = lambda x_: model.f(As, ys, x_) + model.get_reg(x_, reg_name)
            if hasattr(model, 'grad_fx') and model.grad_fx is not None:
                grad_f = lambda x_: model.grad_fx(As, ys, x_)
            else:
                grad_f = lambda x_: jnp.array(jax.grad(lambda z: model.f(As, ys, z))(x_))
        grad = grad_f(x) + lam_gr
        if iter == 0 or len(self.s_list) == 0:
            d = -grad
        else:
            d = self.two_loop_recursion(grad)
        if self.ss_type == 1 and getattr(model, 'L', None) is not None:
            step_size = min(1.0 / model.L, 1.0)
        elif self.ss_type == 1 and getattr(model, 'L', None) is None:
            step_size = 0.5
        elif self.ss_type == 2:
            if iter == 1:
                step_size = 1.0
            else:
                lam_gr_prev = lam * hmu.grad(Cmat, x_prev)
                grad_prev = grad_f(x_prev) + lam_gr_prev
                step_size = inv_BB_step(x, x_prev, grad, grad_prev)
        elif self.ss_type == 3:
            step_size = linesearch(x, d, obj, grad_f)
        else:
            raise ValueError("Please, choose ss_type in [1, 2, 3].")
        Hdiag_inv = 1.0 / (Hr_diag + 1e-9)
        Mg = get_Mg(hmu.Mh, hmu.nu, hmu.mu, len(x))
        eta = np.sqrt(np.dot(lam_gr, lam_gr * Hdiag_inv))
        alpha = step_size / (1 + Mg * eta)
        safe_alpha = min(1, alpha)
        if self.use_prox:
            x_new = prox_ops.invoke_prox(model, reg_name, x + safe_alpha * d, Hdiag_inv, lam, step_size)
        else:
            x_new = x + safe_alpha * d
        grad_fxnew = grad_f(x_new)
        grad_new = grad_fxnew + lam * hmu.grad(Cmat, x_new)
        s = x_new - x
        y = grad_new - grad
        if np.dot(s, y) > 1e-10:
            if len(self.s_list) == self.m:
                self.s_list.pop(0)
                self.y_list.pop(0)
            self.s_list.append(s)
            self.y_list.append(y)
            self.H0 = (np.dot(y, s) / np.dot(y, y)) * np.eye(len(x))
        return x_new, np.linalg.norm(grad_fxnew)



from .exponential_smooth import ExponentialSmootherIndBox
from .log_exp_smooth import LogExpSmootherIndBox
from .phuber_smooth import PHuberSmootherL1L2, PHuberSmootherGL
from .regularizers import *
from .smoothing import *


import numpy as np
import jax.numpy as jnp
import pyscsopt as scs
from pyscsopt.algorithms import ProxGradient, ProxLQNSCORE, ProxGGNSCORE
from pyscsopt.regularizers import PHuberSmootherGL
from pyscsopt.utils import make_group_lasso_problem

np.random.seed(42)

# generate synthetic group lasso regression data using utility
m = 300
n = 100
grpsize = 10
p_active = 0.1  # 10% of groups/features active
A, y, x_true, x0, groups, ind, P = make_group_lasso_problem(
    m=m, n=n, grpsize=grpsize, p_active=p_active, noise_std=0.1, seed=42, group_weights=1.0, use_const_grpsize=True, corr=0.5)

# obj
# define the obj this way ONLY for ProxGGNSCORE (that is, allow for a precomputed yhat, in which case we don't need to provide all the gradient functions below)
def f(A, y, x, yhat=None):
    m = y.shape[0]
    if yhat is None:
        yhat = out_fn(A, x)
    return 0.5 * jnp.sum((yhat - y) ** 2)/m

# strictly OPTIONAL for all methods
def grad_fx(A, y, x):
    m = y.shape[0]
    return (A.T @ (A @ x - y))/m

# the following functions are used ONLY by ProxGGNSCORE (ONLY out_fn is required, others are optional as in other methods provided f is defined as above)
def out_fn(A, x):
    return A @ x

def jac_yx(A, y, yhat, x):
    return A

def grad_fy(A, y, yhat):
    m = y.shape[0]
    return (yhat - y)/m

def hess_fy(A, y, yhat):
    m = y.shape[0]
    return np.eye(len(yhat))/m

# regularization parameters
lam1 = 1e-7  # l1
lam2 = 0.1   # group lasso
lam = [lam1, lam2]
mu = 0.1

x0 = np.random.randn(n)
reg_name = "gl"

# problem = scs.Problem(x0=x0, f=f, lam=lam, A=A, y=y, P=P, out_fn=out_fn, grad_fx=grad_fx, jac_yx=jac_yx, grad_fy=grad_fy, hess_fy=hess_fy)

# the following works fine (gradients are computed internally using jax)
problem = scs.Problem(x0=x0, f=f, lam=lam, A=A, y=y, P=P, out_fn=out_fn)

hmu = PHuberSmootherGL(mu, problem) # group lasso smoother takes also the problem as input

method_pg = ProxGradient(use_prox=True, ss_type=1)
sol_pg = scs.iterate(method_pg, problem, reg_name, hmu, verbose=1, max_epoch=100)

method_lqn = ProxLQNSCORE(use_prox=True, ss_type=2, m=10)
sol_lqn = scs.iterate(method_lqn, problem, reg_name, hmu, verbose=1, max_epoch=100)

method_ggn = ProxGGNSCORE(use_prox=True, ss_type=2)
sol_ggn = scs.iterate(method_ggn, problem, reg_name, hmu, verbose=1, max_epoch=100)

# ### uncomment to print solutions
# print("=" * 50)
# print("True Solution (x_true):")
# print("x_true:", x_true)
# print("=" * 50)
# print("ProxGradient (Sparse Group Lasso):")
# print("Solution x:", sol_pg.x)
# print("=" * 50)
# print("ProxLQNSCORE (Sparse Group Lasso):")
# print("Solution x:", sol_lqn.x)
# print("=" * 50)
# print("ProxGGNSCORE (Sparse Group Lasso):")
# print("Solution x:", sol_ggn.x)



import numpy as np
import jax
import jax.numpy as jnp
from pyscsopt.utils.utils import linesearch, inv_BB_step
from pyscsopt.regularizers.smoothing import get_Mg
import pyscsopt.prox.prox_operators as prox_ops

class ProxGGNSCORE:
    def __init__(self, ss_type=1, use_prox=True, name="prox-ggnscore", label="Prox-GGN-SCORE"):
        self.ss_type = ss_type
        self.use_prox = use_prox
        self.name = name
        self.label = label

    def init(self, x):
        return self

    def ggn_score_step(self, J, Q, gr, Hr_diag, H_inv, residual, lam):
        # gr: list of gradient vectors (each (n,)), or a single (n,) vector
        ncat = len(gr) if isinstance(gr, list) else 1
        n = gr[0].shape[0] if isinstance(gr, list) else gr.shape[0]
        # stack gradients as columns
        if ncat > 1:
            gr_mat = np.column_stack([np.asarray(g).reshape(-1) for g in gr])  # (n, ncat)
        else:
            gr_mat = np.asarray(gr).reshape(n, 1)  # (n, 1)
        # Jt: [J.T  λ*gr_mat]
        Jt = np.hstack([J.T, lam * gr_mat])  # (n, m+ncat)
        # residual: [residual ; ones(ncat)]
        residual_vec = np.concatenate([residual, np.ones(ncat)])
        # Q_aug: block matrix
        qdm1 = Q.shape[0]
        qdm11 = qdm1 + ncat
        Q_aug = np.zeros((qdm11, qdm11))
        Q_aug[:qdm1, :qdm1] = Q
        # the rest is zeros
        if qdm11 <= n:
            A = Q_aug @ (Jt.T @ H_inv) @ Jt
            B = np.linalg.solve(np.eye(qdm11) + A, residual_vec)
            d = H_inv @ Jt @ B
        else:
            JQJ = (Jt @ Q_aug @ Jt.T) + lam * np.diag(Hr_diag)
            Je = Jt @ residual_vec
            JQJ_qr = np.linalg.qr(JQJ)
            JQJ_q, JQJ_r = JQJ_qr[0], JQJ_qr[1]
            d = np.linalg.solve(JQJ_r, JQJ_q.T @ Je)
        return -d

    def step(self, model, reg_name, hmu, As, x, x_prev, ys, Cmat, iter):
        lam = model.lam[0] if hasattr(model.lam, '__len__') and len(model.lam) > 1 else model.lam
        gr = hmu.grad(Cmat, x)
        lam_gr = lam * gr
        Hr_diag = hmu.hess(Cmat, x)
        if all([getattr(model, attr, None) is not None for attr in ("jac_yx", "grad_fy", "hess_fy")]):
            yhat = model.out_fn(As, x)
            J = model.jac_yx(As, ys, yhat, x)
            residual = model.grad_fy(As, ys, yhat)
            Q = model.hess_fy(As, ys, yhat)
        else:
            m_out_fn = lambda z: model.out_fn(As, z)
            yhat = m_out_fn(x)
            f = lambda y_: model.f(As, ys, x, yhat=y_)
            J = jax.jacobian(m_out_fn)(x)
            residual = jax.grad(f)(yhat)
            Q = jax.hessian(f)(yhat)
        if hasattr(model, 'grad_fx') and model.grad_fx is not None:
            grad_f = lambda x_: model.grad_fx(As, ys, x_)
        else:
            grad_f = lambda x_: jnp.array(jax.grad(lambda z: model.f(As, ys, z))(x_))
        Hdiag_inv = 1.0 / (Hr_diag + 1e-9)
        H_inv = np.diag(Hdiag_inv)
        d = self.ggn_score_step(J, Q, [gr], Hr_diag, H_inv, residual, lam)
        if self.ss_type == 1 and getattr(model, 'L', None) is not None:
            step_size = min(1.0 / model.L, 1.0)
        elif self.ss_type == 1 and getattr(model, 'L', None) is None:
            step_size = 0.5
        elif self.ss_type == 2:
            if iter == 1:
                step_size = 1.0
            else:
                lam_gr_prev = lam * hmu.grad(Cmat, x_prev)
                grad_prev = grad_f(x_prev) + lam_gr_prev
                grad_curr = grad_f(x) + lam_gr
                step_size = inv_BB_step(x, x_prev, grad_curr, grad_prev)
        elif self.ss_type == 3:
            step_size = linesearch(x, d, lambda z: model.f(As, ys, z) + model.get_reg(z, reg_name), grad_f)
        else:
            raise ValueError("Please, choose ss_type in [1, 2, 3].")
        Mg = get_Mg(hmu.Mh, hmu.nu, hmu.mu, len(x))
        eta = np.sqrt(np.dot(lam_gr, H_inv @ lam_gr))
        alpha = step_size / (1 + Mg * eta)
        safe_alpha = min(1, alpha)
        if self.use_prox:
            x_new = prox_ops.invoke_prox(model, reg_name, x + safe_alpha * d, Hdiag_inv, lam, step_size)
        else:
            x_new = x + safe_alpha * d
        return x_new, np.linalg.norm(grad_f(x_new))



import numpy as np
import jax
import jax.numpy as jnp
from pyscsopt.utils.utils import linesearch, inv_BB_step
from pyscsopt.regularizers.smoothing import get_Mg
import pyscsopt.prox.prox_operators as prox_ops

class ProxNSCORE:
    def __init__(self, ss_type=1, use_prox=True, name="prox-newtonscore", label="Prox-N-SCORE"):
        self.ss_type = ss_type
        self.use_prox = use_prox
        self.name = name
        self.label = label

    def init(self, x):
        return self

    def step(self, model, reg_name, hmu, As, x, x_prev, ys, Cmat, iter):
        lam = model.lam[0] if hasattr(model.lam, '__len__') and len(model.lam) > 1 else model.lam
        gr = hmu.grad(Cmat, x)
        lam_gr = lam * gr
        Hr_diag = hmu.hess(Cmat, x)
        lam_Hr = lam * np.diag(Hr_diag)
        obj = lambda x_: model.f(As, ys, x_) + model.get_reg(x_, reg_name)
        if hasattr(model, 'grad_fx') and model.grad_fx is not None and hasattr(model, 'hess_fx') and model.hess_fx is not None:
            H = model.hess_fx(As, ys, x)
            grad_f = lambda x_: model.grad_fx(As, ys, x_)
        else:
            H = jnp.array(jax.hessian(obj)(x))
            grad_f = lambda x_: jnp.array(jax.grad(obj)(x_))
        grad = grad_f(x) + lam_gr
        d = -np.linalg.solve(H + lam_Hr, grad)
        if self.ss_type == 1 and getattr(model, 'L', None) is not None:
            step_size = min(1.0 / model.L, 1.0)
        elif self.ss_type == 1 and getattr(model, 'L', None) is None:
            step_size = 0.5
        elif self.ss_type == 2:
            if iter == 1:
                step_size = 1.0
            else:
                lam_gr_prev = lam * hmu.grad(Cmat, x_prev)
                grad_prev = grad_f(x_prev) + lam_gr_prev
                step_size = inv_BB_step(x, x_prev, grad, grad_prev)
        elif self.ss_type == 3:
            step_size = linesearch(x, d, obj, grad_f)
        else:
            raise ValueError("Please, choose ss_type in [1, 2, 3].")
        Hdiag_inv = 1.0 / Hr_diag
        H_inv = np.diag(Hdiag_inv)
        Mg = get_Mg(hmu.Mh, hmu.nu, hmu.mu, len(x))
        d_norm = np.dot(lam_gr, np.dot(H_inv, lam_gr))
        eta = np.sqrt(d_norm)
        alpha = step_size / (1 + Mg * eta)
        safe_alpha = min(1, alpha)
        if self.use_prox:
            x_new = prox_ops.invoke_prox(model, reg_name, x + safe_alpha * d, Hdiag_inv, lam, step_size)
        else:
            x_new = x + safe_alpha * d
        return x_new, np.linalg.norm(grad_f(x_new))


import numpy as np
import jax.numpy as jnp
import pyscsopt as scs
from pyscsopt.algorithms import ProxGradient, ProxLQNSCORE, ProxNSCORE, ProxGGNSCORE
from pyscsopt.regularizers import PHuberSmootherL1L2

np.random.seed(1234)

# Data dimensions
m, n = 100, 300
A = np.random.randn(m, n) * (np.random.rand(m, n) < 1e-2)
x0 = np.random.randn(n)

# true solution (sparse x_true)
x_true = np.zeros(n)
num_nonzero = max(1, n//10)
nonzero_idx = np.random.choice(n, size=num_nonzero, replace=False)
x_true[nonzero_idx] = np.random.randn(num_nonzero)

# generate labels
y_prob = 1 / (1 + jnp.exp(-A @ x_true))
y = np.random.binomial(1, y_prob)
unique_y = np.unique(y)
y = np.where(y == unique_y[0], -1, 1)

# obj
# define the obj this way ONLY for ProxGGNSCORE (that is, allow for a precomputed yhat, in which case we don't need to provide all the gradient functions below)
def f(A, y, x, yhat=None):
    m = y.shape[0]
    if yhat is None:
        yhat = out_fn(A, x)
    logit_yhat = jnp.log(yhat / (1 - yhat))
    return 1/m * jnp.sum(jnp.log(1 + jnp.exp(-y * logit_yhat)))

# the following functions are used ONLY by ProxGGNSCORE (ONLY out_fn is required, others are optional as in other methods provided f is defined as above)
def out_fn(A, x):
    return 1 / (1 + jnp.exp(-A @ x))

def grad_fx(A, y, x):
    Sx = jnp.exp(-y * (A @ x))
    return -A.T @ (y * (Sx / (1 + Sx))) / m

def hess_fx(A, y, x):
    Sx = 1 / (1 + jnp.exp(-y * (A @ x)))
    W = jnp.diag(Sx * (1 - Sx))
    return A.T @ W @ A / m

def jac_yx(A, y, yhat, x):
    return (yhat * (1 - yhat))[:, None] * A

def grad_fy(A, y, yhat):
    return (-y / yhat + (1 - y) / (1 - yhat)) / m

def hess_fy(A, y, yhat):
    return jnp.diag((y / yhat**2 + (1 - y) / (1 - yhat)**2) / m)

reg_name = "l1"
lbda = 1e-2
mu = 1.0
hmu = PHuberSmootherL1L2(mu)

# problem = scs.Problem(x0=x0, f=f, lam=lbda, A=A, y=y, out_fn=out_fn, grad_fx=grad_fx, jac_yx=jac_yx, grad_fy=grad_fy, hess_fy=hess_fy)

# the following works fine (gradients are computed internally using jax)
problem = scs.Problem(x0=x0, f=f, lam=lbda, A=A, y=y, out_fn=out_fn)

method_pg = ProxGradient(use_prox=True, ss_type=1)
sol_pg = scs.iterate(method_pg, problem, reg_name, hmu, verbose=1, max_epoch=100)

method_lqn = ProxLQNSCORE(use_prox=True, ss_type=2, m=10)
sol_lqn = scs.iterate(method_lqn, problem, reg_name, hmu, verbose=1, max_epoch=100)

method_nscore = ProxNSCORE()
sol_nscore = scs.iterate(method_nscore, problem, reg_name, hmu, max_epoch=100, x_tol=1e-6, f_tol=1e-6, verbose=1)

method_ggn = ProxGGNSCORE()
sol_ggn = scs.iterate(method_ggn, problem, reg_name, hmu, max_epoch=100, x_tol=1e-6, f_tol=1e-6, verbose=1)

# ### uncomment to print solutions
# print("=" * 50)
# print("ProxGradient (Sparse Logistic Regression):")
# print("Solution x:", sol_pg.x)
# print("=" * 50)
# print("ProxLQNSCORE (Sparse Logistic Regression):")
# print("Solution x:", sol_lqn.x)
# print("=" * 50)
# print("ProxNSCORE (Sparse Logistic Regression):")
# print("Solution x:", sol_nscore.x)
# print("=" * 50)
# print("ProxGGNSCORE (Sparse Logistic Regression):")
# print("Solution x:", sol_ggn.x)
# print("=" * 50)
# print("True Solution (x_true):")
# print("x_true:", x_true)



import numpy as np

# Proximal operator implementations

def prox_l1(x, h_scale, lam, alpha):
    t = alpha * lam / h_scale
    return np.sign(x) * np.maximum(np.abs(x) - t, 0)

def prox_l2(x, h_scale, lam, alpha):
    t = alpha * lam / h_scale
    denom = np.abs(x)
    denom[denom == 0] = 1e-12
    return x * np.maximum(1 - t/denom, 0)

def prox_indbox(x, model, h_scale, lam, alpha):
    lb, ub = model.C_set[0], model.C_set[1]
    return np.minimum(np.maximum(x, lb), ub)

def prox_grouplasso(x, model, h_scale, lam, alpha):
    P = model.P
    lam1, lam2 = model.lam[0], model.lam[1]
    t = lam1/h_scale
    utmp = np.sign(x) * np.maximum(np.abs(x) - t, 0)  # ProxL1
    u = P.ProxL2(utmp, alpha * lam2, h_scale)  # ProxL2 (assumed implemented in P)
    return u

def invoke_prox(model, reg_name, x, h, lam, alpha):
    if reg_name == "l1":
        return prox_l1(x, h, lam, alpha)
    elif reg_name == "l2":
        return prox_l2(x, h, lam, alpha)
    elif reg_name == "indbox":
        return prox_indbox(x, model, h, lam, alpha)
    elif reg_name == "gl":
        return prox_grouplasso(x, model, h, lam, alpha)
    else:
        raise ValueError("reg_name not valid.")


import numpy as np
from pyscsopt.utils.prox_reg_utils import bounds_sanity_check

def exp_smooth_indbox(x, mu=1.0, lb=0.0, ub=1.0):
    n = x.shape[0]
    a, b = bounds_sanity_check(n, lb, ub)
    return np.exp((-x + a) / mu) * mu

def exp_smooth_grad_indbox(x, mu=1.0, lb=0.0, ub=1.0):
    n = x.shape[0]
    a, b = bounds_sanity_check(n, lb, ub)
    return -np.exp((-x + a) / mu)

def exp_smooth_hess_indbox(x, mu=1.0, lb=0.0, ub=1.0):
    n = x.shape[0]
    a, b = bounds_sanity_check(n, lb, ub)
    return (1 / mu) * np.exp((-x + a) / mu)

class ExponentialSmootherIndBox:
    def __init__(self, C_set, mu):
        self.mu = mu
        self.lb = C_set[0]
        self.ub = C_set[1]
        self.Mh = 1.0
        self.nu = 2.0
    def val(self, x):
        return exp_smooth_indbox(x, self.mu, self.lb, self.ub)
    def grad(self, Cmat, x):
        return exp_smooth_grad_indbox(x, self.mu, self.lb, self.ub)
    def hess(self, Cmat, x):
        return exp_smooth_hess_indbox(x, self.mu, self.lb, self.ub)


import numpy as np
import pyscsopt as scs
from pyscsopt.regularizers import PHuberSmootherL1L2, LogExpSmootherIndBox, ExponentialSmootherIndBox
from pyscsopt.algorithms import ProxGradient, ProxLQNSCORE

import numpy as np
import jax.numpy as jnp
import jax.random as random

np.random.seed(1234)

nvar = 10
Q = jnp.tril(random.uniform(random.PRNGKey(42), (nvar, nvar)))
Q = Q + Q.T - jnp.diag(Q.diagonal())
Q = Q + nvar * jnp.eye(nvar)
c = jnp.ones(nvar)

def f(x):
    return 0.5 * jnp.dot(x, jnp.dot(Q, x)) + jnp.dot(c, x)

x0 = random.uniform(random.PRNGKey(42), (nvar,))
lbda = 0.5
reg_name = "indbox"
mu = 0.1
C_set = (-5.0, 5.0)
hmu = LogExpSmootherIndBox(C_set, mu)
problem = scs.Problem(x0, f, lbda, C_set=C_set)

method_pg = ProxGradient(use_prox=True, ss_type=1)
sol_pg = scs.iterate(method_pg, problem, reg_name, hmu, verbose=1, max_epoch=200)

method_lqn = ProxLQNSCORE(use_prox=True, ss_type=2, m=10)
sol_lqn = scs.iterate(method_lqn, problem, reg_name, hmu, verbose=1, max_epoch=200)

# ### uncomment to print solutions
# print("=" * 50)
# print("ProxGradient:")
# print("Solution x:", sol_pg.x)
# print("=" * 50)
# print("ProxLQNSCORE:")
# print("Solution x:", sol_lqn.x)


import numpy as np
from pyscsopt.utils.prox_reg_utils import bounds_sanity_check

def pseudo_huber(x, mu=1.0):
    return (mu**2 - mu * np.sqrt(mu**2 + x**2) + x**2) / np.sqrt(mu**2 + x**2)

def huber_grad(x, mu=1.0):
    return x / np.sqrt(mu**2 + x**2)

def huber_hess(x, mu=1.0):
    return mu**2/(mu**2 + x**2)**(3/2)

def pseudo_huber_indbox(x, mu=1.0, lb=0.0, ub=1.0):
    n = x.shape[0]
    h = np.zeros(n)
    a, b = bounds_sanity_check(n, lb, ub)
    for i in range(n):
        if x[i] < a[i]:
            h[i] = (-mu * np.sqrt(a[i]**2 - 2*a[i]*x[i] + mu**2 + x[i]**2) + mu**2 + (-x[i] + a[i])**2) / np.sqrt(a[i]**2 - 2*a[i]*x[i] + mu**2 + x[i]**2)
        elif x[i] == a[i] or x[i] <= b[i]:
            h[i] = np.finfo(float).eps
        else:
            h[i] = (-mu * np.sqrt(b[i]**2 - 2*b[i]*x[i] + mu**2 + x[i]**2) + mu**2 + (b[i] - x[i])**2) / np.sqrt(b[i]**2 - 2*b[i]*x[i] + mu**2 + x[i]**2)
    return h

def huber_grad_indbox(x, mu=1.0, lb=0.0, ub=1.0):
    n = x.shape[0]
    g = np.zeros(n)
    a, b = bounds_sanity_check(n, lb, ub)
    for i in range(n):
        if -x[i] < a[i]:
            g[i] = ((a[i]**2 - 2*x[i]*a[i] + mu**2 + x[i]**2)**(-0.5)) * (-x[i] + a[i])
        elif x[i] == a[i] or x[i] < b[i]:
            g[i] = np.finfo(float).eps
        else:
            g[i] = ((b[i]**2 - 2*b[i]*x[i] + mu**2 + x[i]**2)**(-0.5)) * (b[i] - x[i])
    return g

def huber_hess_indbox(x, mu=1.0, lb=0.0, ub=1.0):
    n = x.shape[0]
    h = np.zeros(n)
    a, b = bounds_sanity_check(n, lb, ub)
    for i in range(n):
        if x[i] <= a[i]:
            h[i] = mu**2 * (a[i]**2 - 2*a[i]*x[i] + mu**2 + x[i]**2)**(-1.5)
        elif a[i] < x[i] < b[i]:
            h[i] = np.finfo(float).eps
        elif x[i] >= b[i]:
            h[i] = mu**2 * (b[i]**2 - 2*b[i]*x[i] + mu**2 + x[i]**2)**(-1.5)
    return h

def twonorm(x, kstart, kend):
    return np.linalg.norm(x[kstart:kend+1])

def infconv_huber_norm(x, lam, inds, grpNUM, mu):
    ICz = np.copy(x)
    ind = np.concatenate(inds)
    for j in range(grpNUM):
        lamw = lam * ind[2+3*j]
        kstart = int(ind[3*j])
        kend = int(ind[1+3*j])
        nrm = twonorm(x, kstart, kend)
        for k in range(kstart, kend+1):
            ICz[k] = x[k] * max(1 - lamw / nrm, 0)
            ICz[k] = pseudo_huber(ICz[k], mu)
    return ICz

def get_infconv_huber_l2l1(x, lam1, lam2, inds, grpNUM, mu):
    utmp = infconv_huber_norm(x, lam1, inds, grpNUM, mu)
    z = infconv_huber_norm(utmp, lam2, inds, grpNUM, mu)
    return z

def huber_l2l1_grad(Cmat, x, lam1, lam2, inds, grpNUM, mu):
    g_mu1 = pseudo_huber(x, mu)
    Dg_mu1 = huber_grad(x, mu)
    return huber_grad(np.dot(Cmat, g_mu1), mu) * Dg_mu1

def huber_l2l1_hess(Cmat, x, lam1, lam2, inds, grpNUM, mu):
    g_mu1 = pseudo_huber(x, mu)
    Dg_mu1 = huber_grad(x, mu)
    DDg_mu1 = huber_hess(x, mu)
    DDgg = huber_hess(np.dot(Cmat, g_mu1), mu) * np.dot(Dg_mu1, Dg_mu1) + huber_grad(np.dot(Cmat, g_mu1), mu) * DDg_mu1
    return DDgg

class PHuberSmootherL1L2:
    def __init__(self, mu):
        self.mu = mu
        self.Mh = 2.0
        self.nu = 2.6
    def val(self, x):
        return pseudo_huber(x, self.mu)
    def grad(self, Cmat, x):
        return huber_grad(x, self.mu)
    def hess(self, Cmat, x):
        return huber_hess(x, self.mu)

class PHuberSmootherIndBox:
    def __init__(self, lb, ub, mu):
        self.mu = mu
        self.lb = lb
        self.ub = ub
        self.Mh = 2.0
        self.nu = 2.6
    def val(self, x):
        return pseudo_huber_indbox(x, self.mu, self.lb, self.ub)
    def grad(self, Cmat, x):
        return huber_grad_indbox(x, self.mu, self.lb, self.ub)
    def hess(self, Cmat, x):
        return huber_hess_indbox(x, self.mu, self.lb, self.ub)

class PHuberSmootherGL:
    def __init__(self, mu, model):
        self.mu = mu
        self.Mh = 2.0
        self.nu = 2.6
        lam = model.lam
        P = model.P
        inds = P.ind
        grpNUM = P.grpNUM
        self.lam1 = lam[0]
        self.lam2 = lam[1]
        self.inds = inds
        self.grpNUM = grpNUM
    def val(self, x):
        return get_infconv_huber_l2l1(x, self.lam1, self.lam2, self.inds, self.grpNUM, self.mu)
    def grad(self, Cmat, x):
        return huber_l2l1_grad(Cmat, x, self.lam1, self.lam2, self.inds, self.grpNUM, self.mu)
    def hess(self, Cmat, x):
        return huber_l2l1_hess(Cmat, x, self.lam1, self.lam2, self.inds, self.grpNUM, self.mu)


import numpy as np
import pytest
from pyscsopt.regularizers.phuber_smooth import PHuberSmootherL1L2, PHuberSmootherIndBox, PHuberSmootherGL
from pyscsopt.regularizers.log_exp_smooth import LogExpSmootherIndBox
from pyscsopt.regularizers.exponential_smooth import ExponentialSmootherIndBox

mu = 1
lb = -1.0
ub = 1.0

def test_phuber_l1l2():
    hmu = PHuberSmootherL1L2(mu)
    assert hmu.Mh == 2.0
    assert hmu.nu == 2.6

def test_phuber_indbox():
    hmu = PHuberSmootherIndBox(lb, ub, mu)
    assert hmu.Mh == 2.0
    assert hmu.nu == 2.6

def test_phuber_group_lasso():
    # minimal mock for group lasso model
    class P:
        def __init__(self):
            self.ind = [0, 1]
            self.grpNUM = 1
            self.tau = 0.9
    class Model:
        def __init__(self):
            self.lam = [0.09, 0.91]
            self.P = P()
    model = Model()
    hmu = PHuberSmootherGL(mu, model)
    assert hmu.Mh == 2.0
    assert hmu.nu == 2.6

def test_logexp_smoother_indbox():
    C_set = [lb, ub]
    hmu = LogExpSmootherIndBox(C_set, mu)
    assert hmu.Mh == 1.0
    assert hmu.nu == 2.0

def test_exponential_smoother_indbox():
    C_set = [lb, ub]
    hmu = ExponentialSmootherIndBox(C_set, mu)
    assert hmu.Mh == 1.0
    assert hmu.nu == 2.0



from setuptools import setup, find_packages

setup(
    name="pyscsopt",
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    description="A Python library for self-concordant smooth optimization (Python port of SelfConcordantSmoothOptimization.jl)",
    author="Adeyemi Damilare Adeoye",
    author_email="",
    url="https://github.com/adeyemiadeoye/pyscsopt",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "scipy",
        "jax",
    ],
    python_requires=">=3.8",
    license="Apache-2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)


import numpy as np
import jax
import jax.numpy as jnp
from pyscsopt.utils.utils import linesearch
import pyscsopt.prox.prox_operators as prox_ops

class ProxGradient:
    def __init__(self, ss_type=1, use_prox=True, name="prox-grad", label="Prox-Grad"):
        self.ss_type = ss_type
        self.use_prox = use_prox
        self.name = name
        self.label = label

    def init(self, x):
        return self

    def step(self, model, reg_name, hmu, As, x, x_prev, ys, Cmat, iter):
        # handle regularization parameter
        lam = model.lam[0] if hasattr(model.lam, '__len__') and len(model.lam) > 1 else model.lam
        lam_gr = lam * hmu.grad(Cmat, x)
        Hr_diag = hmu.hess(Cmat, x)
        is_generic = (model.A is None or model.y is None)
        if is_generic:
            obj = lambda x_: model.f(x_) + model.get_reg(x_, reg_name)
            if hasattr(model, 'grad_fx') and model.grad_fx is not None:
                grad_f = lambda x_: model.grad_fx(x_) + lam_gr
            else:
                grad_f = lambda x_: jnp.array(jax.grad(lambda z: model.f(z))(x_)) + lam_gr
        else:
            obj = lambda x_: model.f(As, ys, x_) + model.get_reg(x_, reg_name)
            if hasattr(model, 'grad_fx') and model.grad_fx is not None:
                grad_f = lambda x_: model.grad_fx(As, ys, x_) + lam_gr
            else:
                grad_f = lambda x_: jnp.array(jax.grad(lambda z: model.f(As, ys, z))(x_)) + lam_gr
        grad = grad_f(x)
        d = -grad
        # step size selection
        if self.ss_type == 1 and getattr(model, 'L', None) is not None:
            step_size = 1.0 / model.L
        elif self.ss_type == 1 and getattr(model, 'L', None) is None:
            step_size = linesearch(x, d, obj, grad_f)
        elif self.ss_type == 2:
            # lam_gr_prev = lam * hmu.grad(Cmat, x_prev)
            grad_prev = grad_f(x_prev)
            delta = x - x_prev
            gamma = grad - grad_prev
            L_val = np.dot(gamma, gamma)/np.dot(delta, gamma) if np.dot(delta, gamma) != 0 else 1.0
            step_size = 1.0 if iter == 1 else L_val
        elif self.ss_type == 3:
            step_size = linesearch(x, d, obj, grad_f)
        else:
            step_size = 1.0 / model.L
        if self.use_prox:
            x_new = prox_ops.invoke_prox(model, reg_name, x + step_size * d, np.ones_like(Hr_diag), lam, step_size)
        else:
            x_new = x + step_size * d
        grad_norm = np.linalg.norm(grad)
        return x_new, grad_norm



import numpy as np
import pyscsopt as scs
from pyscsopt.regularizers import PHuberSmootherL1L2
from pyscsopt.algorithms import ProxGradient, ProxLQNSCORE

import numpy as np

np.random.seed(1234)

def f(x):
    return 100 * (x[1] - x[0]**2)**2 + (1 - x[0])**2

x0 = np.array([0.0, -2.0])
lbda = 1e-8 # or set to zero to remove regularization entirely
reg_name = "l1" # see other examples for other regularization names
mu = 1.0
hmu = PHuberSmootherL1L2(mu)
problem = scs.Problem(x0, f, lbda)

# proximal gradient
method_pg = ProxGradient(use_prox=True, ss_type=1)
sol_pg = scs.iterate(method_pg, problem, reg_name, hmu, verbose=1, max_epoch=100)

# proximal LQNSCORE (L-BFGS-SCORE)
method_lqn = ProxLQNSCORE(use_prox=True, ss_type=2, m=10)
sol_lqn = scs.iterate(method_lqn, problem, reg_name, hmu, verbose=1, max_epoch=100)

# ### uncomment to print solutions
# print("=" * 50)
# print("ProxGradient:")
# print("Solution x:", sol_pg.x)
# print("=" * 50)
# print("ProxLQNSCORE:")
# print("Solution x:", sol_lqn.x)